@@ -1,6 +1,6 @@
 /* 
    sitecopy, for managing remote web sites.
-   Copyright (C) 1998-2002, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1998-2004, Joe Orton <joe@manyfish.co.uk>
                                                                      
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -44,6 +44,11 @@
 #ifdef HAVE_ERRNO_H
 #include <errno.h>
 #endif
+
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
+
 #include <signal.h>
 #include <time.h>
 #include <ctype.h>
@@ -151,13 +156,13 @@ static int allsites, /* Do they want all sites to be operated on? */
     listflat, /* Do they want the 'flat' list style */
     show_progress, /* Do they want the %-complete messages */
     prompting, /* Did they say --prompting? */
-    keepgoing; /* Did they say --keep-going? */
+    keepgoing, /* Did they say --keep-going? */
+    dry_run;   /* Did they say --dry-run? */
 
 /* Functions prototypes */
 static void init(int, char **);
 static void init_signals(void);
 static void parse_cmdline(int, char **);
-static int issue_error(struct site *site, enum action act, int error);
 static int act_on_site(struct site *site, enum action act);
 static int verify_sites(struct site *sites, enum action act);
 int main(int, char **);
@@ -172,10 +177,13 @@ int main(int argc, char *argv[])
     int ret = 0, numgoodsites;
     struct site *current;
 
-#ifdef ENABLE_NLS
+#ifdef HAVE_SETLOCALE
     setlocale(LC_ALL, "");
-    bindtextdomain(PACKAGE, LOCALEDIR);
-    textdomain(PACKAGE);
+#endif
+
+#ifdef ENABLE_NLS
+    bindtextdomain(PACKAGE_NAME, LOCALEDIR);
+    textdomain(PACKAGE_NAME);
 #endif /* ENABLE_NLS */
 
     fe_initialize();
@@ -451,7 +459,7 @@ static void parse_cmdline(int argc, char *argv[])
 #ifdef NE_DEBUGGING
 	"d:g:"
 #endif
-	"acefhiklop:qr:suvVyZ"; /* available: bgjmntwx */
+	"acefhiklonp:qr:suvVyZ"; /* available: bgjmntwx */
     const static struct option longopts[] = {
 	/* Operation modes */
 	{ "update", no_argument, NULL, 'u' },
@@ -462,6 +470,7 @@ static void parse_cmdline(int argc, char *argv[])
 	{ "list", no_argument, NULL, 'l' },
 	{ "flatlist", no_argument, NULL, 'Z' },
 	{ "keep-going", no_argument, NULL, 'k' },
+        { "dry-run", no_argument, NULL, 'n' },
 	{ "show-progress", no_argument, NULL, 'o' },
 /*	{ "force-overwrite", no_argument, NULL, 't' }, */
 	{ "help", no_argument, NULL, 'h' },
@@ -580,6 +589,9 @@ static void parse_cmdline(int argc, char *argv[])
 	case 'a':
 	    allsites = true;
 	    break;
+        case 'n':
+            dry_run = true;
+            break;
 	case 'o':
 	    show_progress = true;
 	    break;
@@ -591,7 +603,7 @@ static void parse_cmdline(int argc, char *argv[])
 	case '?': 
 	default:
 	    printf(_("Try `%s --help' for more information.\n"), progname);
-	    exit(-1);
+	    exit(1);
 	}
     }
 
@@ -599,6 +611,14 @@ static void parse_cmdline(int argc, char *argv[])
     if (action == action_none)
 	action = action_list;
 
+    /* Dry-run mode is only currently supported for --update */
+    if (dry_run && action != action_update) {
+        printf(_("%s: Error: Dry run mode is currently only supported "
+                 "for updates.\n"),
+               progname);
+        exit(1);
+    }
+
 #ifdef NE_DEBUGGING
     ne_debug_init(use_debug_stream, use_debug_mask);
 #else
@@ -627,6 +647,35 @@ static int get_username(const char *prompt, char *buffer)
     }
 }
 
+int fe_accept_cert(const ne_ssl_certificate *cert, int failures)
+{
+    const char *id = ne_ssl_cert_identity(cert);
+    char *dn, fprint[61];
+
+    ne_ssl_cert_digest(cert, fprint);
+
+    puts(_("WARNING: Server certificate is not trusted."));
+    
+    if (id)
+        printf(_("Certificate was issued for server `%s'.\n"), id);
+    else
+        puts(_("WARNING: Certificate does not specify a server"));
+    
+    printf(_("Fingerprint: %s\n"), fprint);
+
+    dn = ne_ssl_readable_dname(ne_ssl_cert_subject(cert));
+    printf(_("Issued to: %s\n"), dn);
+    free(dn);
+
+    dn = ne_ssl_readable_dname(ne_ssl_cert_issuer(cert));
+    printf(_("Issued by: %s\n"), dn);
+    free(dn);
+
+    printf(_("Do you wish to accept this certificate? (y/n) "));
+
+    return !yesno();
+}
+
 int fe_login(fe_login_context ctx, const char *realm, const char *hostname,
 	     char *username, char *password) 
 {
@@ -755,7 +804,8 @@ void fe_checksummed(const char *file, int success, const char *err)
 void fe_setting_perms(const struct site_file *file)
 {
     if (quiet == 0) {
-	printf(_("Setting permissions on %s: "), file_name(file));
+	printf(_("Setting permissions on %s%s: "), file_name(file),
+               file->type == file_dir ? "/" : "");
 	fflush(stdout);
     }
 }
@@ -829,10 +879,13 @@ void fe_updating(const struct site_file *file)
 
 void fe_updated(const struct site_file *file, int success, const char *error)
 {
+    char wrap = error && strlen(error) < 30 ? ' ' : '\n';
+
     upload_sofar += file->local.size;
+
     if (quiet > 0) {
 	if (! success) {
-	    printf(_("Failed to update %s:\n%s\n"), file_name(file), error);
+	    printf(_("Failed to update %s:%c%s\n"), file_name(file), wrap, error);
 	}
 	return;
     }
@@ -841,7 +894,7 @@ void fe_updated(const struct site_file *file, int success, const char *error)
 	if (success) {
 	    printf(_("done.\n"));
 	} else {
-	    printf(_("failed:\n%s\n"), error);
+	    printf(_("failed:%c%s\n"), wrap, error);
 	}
     } else {
 	if (success) {
@@ -853,7 +906,7 @@ void fe_updated(const struct site_file *file, int success, const char *error)
 		printf(_("] done.\n"));
 	    }
 	} else {
-	    printf(_("] failed:\n%s\n"), error);
+	    printf(_("] failed:%c%s\n"), wrap, error);
 	}
     }
 }
@@ -1214,7 +1267,9 @@ static void init_signals(void)
      * site update, the state gets saved properly. */
     signal(SIGINT, abort_handler);
     signal(SIGHUP, abort_handler);
+#if 0
     signal(SIGSEGV, abort_handler);
+#endif
     signal(SIGTERM, abort_handler);
     signal(SIGUSR1, abort_handler);
     signal(SIGUSR2, abort_handler);
@@ -1230,7 +1285,6 @@ static int act_on_site(struct site *site, enum action act)
 
     /* Set the options */
     site->keep_going = keepgoing;
-    site->only_marked = false; /* TODO */
 
     switch (act) {
     case action_update:
@@ -1239,7 +1293,11 @@ static int act_on_site(struct site *site, enum action act)
 		printf(_("%s: Nothing to do - no changes found.\n"), 
 		       progname);
 	    }
-	} else {
+	} else if (dry_run) {
+            extern const struct proto_driver null_driver;
+            site->driver = &null_driver;
+            ret = issue_error(site, act, site_update(site));
+        } else {
 	    if (site_open_storage_file(site) == NULL) {
 		printf(_("%s: Error: Could not open storage file for writing (%s)\n"
 			  "%s: Error: %s\n"
@@ -1360,7 +1418,7 @@ static void usage(void)
 #ifdef NE_DEBUGGING
     printf(_(
 "  -d, --debug=KEY[,KEY] Turn debugging on for each KEY, which may be:\n"
-"     socket, files, rcfile, ftp, http, httpbody, xml, xmlparse, cleartext\n"
+"     socket, files, rcfile, ftp, http, httpbody, rsh, sftp, xml, xmlparse, cleartext\n"
 "     Warning: cleartext displays (normally hidden) passwords in plain text\n"
 "  -g, --logfile=FILE    Append debugging messages to FILE (else use stderr)\n"
 ));
@@ -1374,6 +1432,7 @@ static void usage(void)
 "  -o, --show-progress   Display total percentage file transfer complete\n"
 "  -q, --quiet           Be quiet while performing the operation\n"
 "  -qq, --silent         Be silent while perforing the operation\n"
+"  -n, --dry-run         Display but do not carry out the operation\n"
 "Operation modes:\n"
 "  -l, --list            List changes between remote and local sites (default)\n"
 "  -ll, --flatlist       Flat list of changes between remote and local sites\n"
@@ -1393,7 +1452,7 @@ static void usage(void)
 /* Two-liner version information. */
 static void version(void) 
 {
-    printf(PACKAGE " " VERSION ":");
+    printf(PACKAGE_NAME " " PACKAGE_VERSION ":");
 #ifdef USE_FTP
     printf(" FTP");
 #endif /* FTP */
@@ -1403,6 +1462,9 @@ static void version(void)
 #ifdef USE_RSH
     printf(" rsh/rcp");
 #endif /* USE_RSH */
+#ifdef USE_SFTP
+    printf(" sftp/ssh");
+#endif /* USE_SFTP */
 #ifdef NE_DEBUGGING
     printf(", debugging");
 #endif
