@@ -339,8 +339,11 @@ static int parse_reply(ftp_session *sess, int code, char *reply)
     case 221:
 	/* succesful QUIT reseponse. */
 	return FTP_CLOSED;
-    case 421: /* service denied */
-	return FTP_DENIED;
+    case 421: /* service denied; PI connection closure */
+        ne_sock_close(sess->pisock);
+        sess->pisock = NULL;
+        sess->connected = 0;
+        return FTP_BROKEN;
     case 553: /* couldn't create directory */
 	return FTP_ERROR;
     case 213: /* MDTM response, hopefully */
@@ -643,6 +646,7 @@ static int dtp_open_active(ftp_session *sess, const char *command)
 	int errnum = errno;
 	set_syserr(sess, _("Active open failed: could not determine "
 			   "address of control socket"), errnum);
+        return FTP_ERROR;
     }
 
     /* Let the kernel choose a port */
@@ -676,6 +680,13 @@ static int dtp_open_active(ftp_session *sess, const char *command)
 	return FTP_ERROR;
     }
 
+    if (addr.sin_port == 0) {
+        ftp_seterror(sess, _("Could not determine bound port number for "
+                             "data socket"));
+        close(listener);
+        return FTP_ERROR;
+    }
+
     if (listen(listener, 1) < 0) {
 	int errnum = errno;
 	set_syserr(sess, ("Active open failed: could not listen for "
@@ -1053,8 +1064,10 @@ int ftp_open(ftp_session *sess)
     }
 
     if (authenticate(sess) != FTP_OK) {
-        ne_sock_close(sess->pisock);
-        sess->pisock = NULL;
+        if (sess->connected) {
+            ne_sock_close(sess->pisock);
+            sess->pisock = NULL;
+        }
 	return FTP_LOGIN;
     }
 
@@ -1084,22 +1097,13 @@ const char *ftp_get_error(ftp_session *sess)
     return sess->error;
 }    
 
-/* This does the ls -laR, and tries it's best to parse the resulting
- * list. Currently implemented only for Unix-style servers, which go:
- * dirlist
- * new/directory/name:
- * dirlist
- * another/directory/name:
- * dirlist
- * etc. where dirlist is a straight ls -al listing
- */
 int ftp_fetch(ftp_session *sess, const char *startdir, struct proto_file **list)
 {
     struct proto_file *tail = NULL;
     struct ls_context *lsctx;
     int ret;
 
-    if ((ret = ftp_data_open(sess, "LIST -laR %s", startdir)) != FTP_READY) {
+    if ((ret = ftp_data_open(sess, "LIST -la %s", startdir)) != FTP_READY) {
         return FTP_ERROR;
     }
 
@@ -1126,8 +1130,12 @@ int ftp_fetch(ftp_session *sess, const char *startdir, struct proto_file **list)
 
         lsrv = ls_parse(lsctx, sess->rbuf, &lfile);
         if (lsrv == ls_error) {
+            char err[512];
+
+            ne_snprintf(err, sizeof err, _("Parse error in LIST response: %s"),
+                        ls_geterror(lsctx));
             NE_DEBUG(DEBUG_FTP, "ftp: ls_parse error, aborting.\n");
-            ftp_seterror(sess, _("Parse error in LIST response"));
+            ftp_seterror(sess, err);
             ret = FTP_ERROR;
         } else {
             ls_pflist_add(list, &tail, &lfile, lsrv);
