@@ -1,6 +1,6 @@
 /* 
    sitecopy, for managing remote web sites.
-   Copyright (C) 1998-2001, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1998-2004, Joe Orton <joe@manyfish.co.uk>
                                                                      
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -61,8 +61,8 @@ char *rcfile_err;
 /** Not quite so global variables **/
 
 /* These are appended to $HOME */
-#define RCNAME "/." PACKAGE "rc"
-#define COPYNAME "/." PACKAGE "/"
+#define RCNAME "/.sitecopyrc"
+#define COPYNAME "/.sitecopy/"
 #define NETRCNAME "/.netrc"
 
 /* Stores the list of entries in the ~/.netrc */
@@ -80,6 +80,9 @@ extern const struct proto_driver dav_driver;
 #ifdef USE_RSH
 extern const struct proto_driver rsh_driver;
 #endif /* USE_RSH */
+#ifdef USE_SFTP
+extern const struct proto_driver sftp_driver;
+#endif /* USE_SFTP */
 
 /* rcfile_read will read the rcfile and fill given sites list.
  * This returns 0 on success, RC_OPENFILE if the rcfile could not
@@ -107,7 +110,10 @@ int rcfile_read(struct site **sites)
     default_site.perms = sitep_ignore;
     default_site.symlinks = sitesym_follow;
     default_site.protocol = siteproto_ftp;
+    default_site.proto_string = ne_strdup("ftp");
+
     default_site.ftp_pasv_mode = true;
+    default_site.ftp_use_cwd = false;
     
     last_site = this_site = NULL;
     rcfile_linenum = 0;
@@ -269,8 +275,10 @@ int rcfile_read(struct site **sites)
 		}		
 		this_site->name = ne_strdup(val);
 		this_site->files = NULL;
+		this_site->proto_string = ne_strdup(default_site.proto_string);
 		/* Now work out the info filename */
-		CONCAT2(this_site->infofile, copypath, val);
+		this_site->infofile = ne_concat(copypath, val, NULL);
+                this_site->certfile = ne_concat(copypath, val, ".crt", NULL);
 	    } else if (this_site == NULL) {
 		ret = RC_CORRUPT;
 	    } else if (strcmp(key, "username") == 0) {
@@ -325,10 +333,13 @@ int rcfile_read(struct site **sites)
 	    } else if (strcmp(key, "permissions") == 0) {
 		if (strcmp(val, "ignore") == 0) {
 		    this_site->perms = sitep_ignore;
+                    this_site->dirperms = 0;
 		} else if (strcmp(val, "exec") == 0) {
 		    this_site->perms = sitep_exec;
 		} else if (strcmp(val, "all") == 0) {
 		    this_site->perms = sitep_all;
+                } else if (strcmp(val, "dir") == 0) {
+                    this_site->dirperms = 1;
 		} else {
 		    ret = RC_CORRUPT;
 		}
@@ -378,15 +389,28 @@ int rcfile_read(struct site **sites)
 		    this_site->protocol = siteproto_dav;
 		} else if (strcasecmp(val, "rsh") == 0) {
 		    this_site->protocol = siteproto_rsh;
-		} else {
+		} else if (strcasecmp(val, "ssh") == 0) {
+                    this_site->protocol = siteproto_rsh;
+                    if (this_site->rsh_cmd == NULL) 
+                        this_site->rsh_cmd = ne_strdup("ssh");
+                    if (this_site->rcp_cmd == NULL) 
+                        this_site->rcp_cmd = ne_strdup("scp");
+		} else if (strcasecmp(val, "sftp") == 0) {
+		    this_site->protocol = siteproto_sftp;
+                } else {
 		    this_site->protocol = siteproto_unknown;
 		}
+		free(this_site->proto_string);
 		this_site->proto_string = ne_strdup(val);
 	    } else if (strcmp(key, "ftp") == 0) {
 		if (strcmp(val, "nopasv") == 0) {
 		    this_site->ftp_pasv_mode = false;
 		} else if (strcmp(val, "showquit") == 0) {
 		    this_site->ftp_echo_quit = true;		    
+		} else if (strcmp(val, "usecwd") == 0) {
+		    this_site->ftp_use_cwd = true;		    
+		} else if (strcmp(val, "nousecwd") == 0) {
+		    this_site->ftp_use_cwd = false;		    
 		} else {
 		    ret = RC_CORRUPT;
 		}
@@ -421,7 +445,9 @@ int rcfile_read(struct site **sites)
 		} else {
 		    ret = RC_CORRUPT;
 		}
-	    } else {
+	    } else if (strcmp(key, "charset") == 0) {
+                NE_DEBUG(DEBUG_RCFILE, "Ignored key %s\n", key);
+            } else {
 		/* Unknown key! */
 		ret = RC_CORRUPT;
 	    }
@@ -478,7 +504,7 @@ int rcfile_verify(struct site *any_site)
 	if (any_site->remote_isrel) { 
 	    return SITE_NOREMOTEREL;
 	}
-	if (any_site->perms == sitep_all) {
+	if (any_site->perms == sitep_all || any_site->dirperms) {
 	    return SITE_NOPERMS;
 	}
 	if (any_site->symlinks == sitesym_maintain) {
@@ -496,6 +522,14 @@ int rcfile_verify(struct site *any_site)
 #else /* !USE_RSH */
 	return SITE_UNSUPPORTED;
 #endif /* USE_RSH */
+    case siteproto_sftp:
+#ifdef USE_SFTP
+	any_site->driver = &sftp_driver;
+	/* FIXME: sftp checks? */
+	break;
+#else /* !USE_SFTP */
+	return SITE_UNSUPPORTED;
+#endif /* USE_SFTP */
     case siteproto_unknown:
 	return SITE_UNSUPPORTED;
     }
@@ -551,7 +585,8 @@ int rcfile_verify(struct site *any_site)
     }
     if (any_site->local_isrel) {
 	/* We skip the first char ('~') of l_r_u */
-	CONCAT2(any_site->local_root, home, any_site->local_root_user + 1);
+	any_site->local_root = ne_concat(home, any_site->local_root_user + 1,
+					 NULL);
     } else {
 	any_site->local_root = any_site->local_root_user;
     }
@@ -559,7 +594,7 @@ int rcfile_verify(struct site *any_site)
     /* Now check the local directory actually exists.
      * To do this, stat `/the/local/root/.', which will fail if the
      * can't read the directory or if it's a file not a directory */
-    CONCAT2(temp, any_site->local_root, ".");
+    temp = ne_concat(any_site->local_root, ".", NULL);
     ret = stat(temp, &localst);
     free(temp);
     if (ret != 0) {
@@ -596,7 +631,8 @@ int init_netrc() {
 }
 
 /* Checks the perms of the rcfile and site storage directory. */
-int init_paths() {
+int init_paths()
+{
     struct stat st;
     if (stat(rcfile, &st) < 0) {
 	NE_DEBUG(DEBUG_RCFILE, "stat failed on %s: %s\n", 
@@ -604,6 +640,9 @@ int init_paths() {
 	return RC_OPENFILE;
     }
 #if !defined (__EMX__) && !defined(__CYGWIN__)
+    if (!S_ISREG(st.st_mode)) {
+        return RC_OPENFILE;
+    }
     if ((st.st_mode & ~(S_IFREG | S_IREAD | S_IWRITE)) > 0) {
 	return RC_PERMS;
     }
@@ -623,7 +662,7 @@ int init_paths() {
 	return RC_DIROPEN;
     }
 #if !defined (__EMX__) && !defined(__CYGWIN__)
-    if ((st.st_mode & ~(S_IFDIR | S_IREAD | S_IWRITE | S_IEXEC)) > 0) {
+    if (st.st_mode & (S_IRWXG | S_IRWXO)) {
 	return RC_DIRPERMS;
     }
 #endif
@@ -643,12 +682,12 @@ int init_env() {
 	}
     }
     if (rcfile == NULL) {
-	CONCAT2(rcfile, home, RCNAME);
+	rcfile = ne_concat(home, RCNAME, NULL);
     }
     if (copypath == NULL) {
-	CONCAT2(copypath, home, COPYNAME);
+	copypath = ne_concat(home, COPYNAME, NULL);
     }
-    CONCAT2(netrcfile, home, NETRCNAME);
+    netrcfile = ne_concat(home, NETRCNAME, NULL);
     return 0;
 }
 
@@ -707,8 +746,7 @@ int rcfile_write (char *filename, struct site *list_of_sites)
 	    return RC_CORRUPT;
 	}
 	
-	if (fprintf (fp, "  protocol %s\n", 
-		     current->protocol==siteproto_ftp?"ftp":"http") == -1) {
+	if (fprintf (fp, "  protocol %s\n", current->proto_string) == -1) {
 	    return RC_CORRUPT;
 	}
 	
@@ -740,6 +778,7 @@ int rcfile_write (char *filename, struct site *list_of_sites)
 	
 	RCWRITEBOOL(!current->ftp_pasv_mode, "ftp nopasv");
 	RCWRITEBOOL(current->ftp_echo_quit, "ftp showquit");
+	RCWRITEBOOL(current->ftp_use_cwd, "ftp usecwd");
 	RCWRITEBOOL(current->http_limit, "http limit");
 	RCWRITEBOOL(current->http_use_expect, "http expect");
 	
@@ -791,6 +830,12 @@ int rcfile_write (char *filename, struct site *list_of_sites)
 	    break;
 	}
 	
+        if (current->dirperms) {
+            if (fprintf(fp, "  permissions dir\n") == -1) {
+                return RC_CORRUPT;
+            }
+        }
+
 	/* Sym link mode */
 	switch (current->symlinks) {
 	case (sitesym_ignore): 
